// DTMNT emu - config file

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <allegro.h>
#include "dtmnt.h"

unsigned char dip[4]={0,0x00,0xa2,0x00};

int *scode_remap=NULL;
double alter_pitch=1.0;

int rom_set=1;
extern int ym2151_amp,drum_volume;
int record_playback=0;
int stop_flicker=1;
extern int force_dsp_version;
extern int using_digi_card,stereo16bit;
extern int dodgy_drums;
extern int reboot,save_state;
extern int sbsnd_base,irq_num,dma_chan_8,dma_chan_16;
extern int get_from_blaster;
extern int show_sound_info;
extern int lpt_read;

char driver_name[32][32]=
{
"GFX_TEXT",
"GFX_AUTODETECT",
"GFX_VGA",
"GFX_MODEX",
"GFX_VESA1",
"GFX_VESA2Banked",
"GFX_VESA2Linear",
"GFX_VESA3",
"GFX_VBEAF",
"GFX_XTENDED",
"GFX_ATI",
"GFX_MACH64",
"GFX_CIRRUS64",
"GFX_CIRRUS54",
"GFX_PARADISE",
"GFX_S3",
"GFX_TRIDENT",
"GFX_ET3000",
"GFX_ET4000",
"GFX_ET6000",
"GFX_VIDEO7"
""
};

int command_line_options(int argc,char **argv)
{
  char *arg; int i;
  for (i=1;i<argc;i++)
  {
    arg=argv[i];
    switch (*arg)
    {
      case '-': case '/':
        if (strcasecmp(arg+1,"record")==0) record_playback=1;
        else if (strcasecmp(arg+1,"playback")==0) record_playback=-1;
        else if (strcasecmp(arg+1,"reboot")==0) reboot=1;
        else if (strcasecmp(arg+1,"nosound")==0) want_sound=0;
        else if (strcasecmp(arg+1,"sound")==0) want_sound=1;
        else if (strcasecmp(arg+1,"vesa2l")==0)
        { gr.driver=GFX_VESA2L; gr.x=320; gr.y=240; gr.depth=16; }

        else if (strcasecmp(arg+1,"vesa2b")==0)
        { gr.driver=GFX_VESA2B; gr.x=320; gr.y=240; gr.depth=16; }

        else if (strcasecmp(arg+1,"modex")==0)
        { gr.driver=GFX_MODEX; gr.x=320; gr.y=240; gr.depth=8; }
        else if (strcasecmp(arg+1,"vga")==0)
        { gr.driver=GFX_VGA; gr.x=320; gr.y=200; gr.depth=8; }
        else if (strcasecmp(arg+1,"vesa1")==0)
        { gr.driver=GFX_VESA1; gr.x=640; gr.y=480; gr.depth=16; }
        else if (strcasecmp(arg+1,"subtle")==0)
        { gr.driver=GFX_TEXT; gr.depth=8; }
        else if (strncasecmp(arg+1,"res",3)==0)
          sscanf(arg+4,"%dx%dx%d",&gr.x,&gr.y,&gr.depth);

      break;
      default:
      strncpy(rom_path,argv[1],255);
    }
  }
  return 0;
}

// parses the config file

int load_config(char *name)
{
  int done=0;
  FILE *hand;
  int i;
  int scode_remapped=0;
  int no_config_file=0;

  // First of all init the sound code remaps
  scode_remap=tralloc(256*sizeof(int));
  if (scode_remap!=NULL)
  {
    for (i=0;i<256;i++) scode_remap[i]=i;
  }

  hand=fopen(name,"rt");
  if (hand==NULL)
  {
    no_config_file=1;
  }
  else
  {
    no_config_file=0;
    printf ("    <--- Getting config from '%s'\n",name);
  
    done=0;
    while (!done)
    {
      static char buf[256]="";
      if (fgets(buf,255,hand)==NULL) done=1;
      else
      {
        char *check_for=NULL;
        int slen=strlen(buf);

        // Knock off linefeeds, because they will mess up strings
        if (buf[slen-1]==10) {buf[slen-1]=0; slen--;}

#define IT_IS (strncasecmp(buf,check_for,strlen(check_for))==0)
#define VALUE (buf+strlen(check_for))

        check_for="rom_path:";
        if (IT_IS) strncpy(rom_path,VALUE,255);

        if (strncasecmp(buf,"gr.driver",9)==0)
          gr.driver=atoi(buf+9);
        else if (strncasecmp(buf,"gr.x",4)==0)
          gr.x=atoi(buf+4);
        else if (strncasecmp(buf,"gr.y",4)==0)
          gr.y=atoi(buf+4);
        else if (strncasecmp(buf,"gr.depth",8)==0)
          gr.depth=atoi(buf+8);
        else if (strncasecmp(buf,"layer",5)==0)
          layer=strtol(buf+5,NULL,0);
        else if (strncasecmp(buf,"wait_vbl",8)==0)
          wait_vbl=atoi(buf+8);
        else if (strncasecmp(buf,"auto_frameskip",14)==0)
          auto_frameskip=atoi(buf+14);
        else if (strncasecmp(buf,"virtual_fps",11)==0)
          virtual_fps=atoi(buf+11);
        else if (strncasecmp(buf,"frame_skip",10)==0)
          frame_skip=atoi(buf+10);
        else if (strncasecmp(buf,"dip_switch",10)==0)
        {
          int no=buf[10]-'0';
          no&=3;
          dip[no]=strtol(buf+11,NULL,0);
        }
        else if (strncasecmp(buf,"fit_screen",10)==0)
          fit_screen=atoi(buf+10);
        else if (strncasecmp(buf,"want_sound",10)==0)
          want_sound=atoi(buf+10);
        else
        {
          check_for="emulator_keys_off";
          if (IT_IS) emulator_keys_off=atoi(VALUE);
  
          // tricky bit for checking for input_type[p][k]
          check_for="input_type[";
          if (IT_IS)
          {
            char *next_bit;
            int p=strtol(VALUE,&next_bit,0),k=0;
            if (next_bit!=NULL)
            {
              if (next_bit[0]==']')
              {
                if (next_bit[1]=='[')
                {
                  k=strtol(next_bit+2,&next_bit,0);
                  if (next_bit!=NULL)
                  {
                    if (next_bit[0]==']')
                    {
                      if (p<0) p=0; else if (p>4) p=4;
                      if (k<0) k=0; else if (k>7) k=7;
  
                      input_type[p][k]=strtol(next_bit+2,NULL,0);
                      // phew!
                    }
                  }
                }
              }
            }
          }
  
          check_for="set_joy_type";
          if (IT_IS) set_joy_type=atoi(VALUE);
          check_for="show_fps";
          if (IT_IS) show_fps=atoi(VALUE);
  
          check_for="waveout_size";
          if (IT_IS) waveout_size=atoi(VALUE);
          check_for="waveout_rate";
          if (IT_IS) waveout_rate=atoi(VALUE);
          check_for="timerb_speed";
          if (IT_IS) timerb_speed=atoi(VALUE);
  
          check_for="scode_remapped";
          if (IT_IS) scode_remapped=1;
          else
          {
            check_for="scode_remap";
            if (IT_IS)
            {
              int from=0,to=0; char *next=NULL;
              from=strtol(VALUE,&next,0);
              if (next!=NULL)
                to=strtol(next,NULL,0);
              if (scode_remap!=NULL)
              {
                printf ("> Remapping sound 0x%.2x to 0x%.2x\r",from,to);
                fflush(stdout);
                scode_remap[from&0xff]=to;
              }
            }
          }

          check_for="display_scodes";
          if (IT_IS) display_scodes=atoi(VALUE);
          check_for="gamma_correct";
          if (IT_IS) gamma_correct=atoi(VALUE);
          check_for="gamma_value";
          if (IT_IS) gamma_value=atof(VALUE);
          check_for="trippy_sprites";
          if (IT_IS) trippy_sprites=atoi(VALUE);
          check_for="ym2151_amp";
          if (IT_IS) ym2151_amp=atoi(VALUE);
          check_for="drum_volume";
          if (IT_IS) drum_volume=atoi(VALUE);
          check_for="alter_pitch";
          if (IT_IS) alter_pitch=atof(VALUE);
          check_for="force_dsp_version";
          if (IT_IS) force_dsp_version=strtol(VALUE,NULL,0);

          check_for="record_playback";
          if (IT_IS) record_playback=atoi(VALUE);
          check_for="rom_set";
          if (IT_IS) rom_set=atoi(VALUE);
          check_for="stop_flicker";
          if (IT_IS) stop_flicker=atoi(VALUE);

          check_for="using_digi_card";
          if (IT_IS) using_digi_card=atoi(VALUE);
          check_for="dodgy_drums";
          if (IT_IS) dodgy_drums=atoi(VALUE);

          check_for="reboot";
          if (IT_IS) reboot=atoi(VALUE);
          check_for="save_state";
          if (IT_IS) save_state=atoi(VALUE);

          check_for="sbsnd_base";
          if (IT_IS) sbsnd_base=strtol(VALUE,NULL,16);
          check_for="irq_num";
          if (IT_IS) irq_num=atoi(VALUE);
          check_for="dma_chan_8";
          if (IT_IS) dma_chan_8=atoi(VALUE);
          check_for="dma_chan_16";
          if (IT_IS) dma_chan_16=atoi(VALUE);
          check_for="stereo16bit";
          if (IT_IS) stereo16bit=atoi(VALUE);
          check_for="get_from_blaster";
          if (IT_IS) get_from_blaster=atoi(VALUE);
          check_for="show_sound_info";
          if (IT_IS) show_sound_info=atoi(VALUE);
          check_for="lpt_read";
          if (IT_IS) lpt_read=atoi(VALUE);
        }
      }
    }
  
    fclose(hand);
  }

  printf ("\n");
  // If we didn't have a sound code remap - do some defaults
  // (get rid of nasty sounds)
  // turn off (or change) painful effects!
  if (!scode_remapped)
    if (scode_remap!=NULL)
    {
      printf ("Doing default scode_remaps\n");
      // New ym2151 handles them better
      /*
      scode_remap[0x04]=0x10;
      scode_remap[0x05]=0x0d; // hitting a robot!
      scode_remap[0x1d]=0x23;
      scode_remap[0x25]=0x0d; // Hitting shuriken
      scode_remap[0x37]=0x08;
      //scode_remap[0x40]=0x17; smashing a window
      */
      // (nice - but loud...!)
    }

  return no_config_file;
}

int unload_config(char *name)
{
  int i;
  FILE *hand=fopen(name,"wt");

  if (hand==NULL) return 1;
  printf ("  ----> Saving config to '%s'\n",name);

  fprintf(hand,"// Dave's TMNT v%s config file\n\n",VER);
  fprintf(hand,"// Note, 0=off, 1=on\n\n");

  fprintf(hand,"// PLEASE READ THE ENTIRE FILE, to see if you want to change anything\n");
  fprintf(hand,"// ------------------------------------------------------------------\n\n");

  fprintf(hand,"// Set this to 1 to reboot machine (needed if you change dip_switches\n");
  fprintf(hand,"reboot 0\n");
  fprintf(hand,"save_state %d\n",save_state);

  fprintf(hand,"//(set next varibale to 1 to record, -1 to playback\n");
  fprintf(hand,"//ONLY AFFECTS NEXT GAME and resets to 0 after)\n\n");
  fprintf(hand,"record_playback 0\n\n");

  fprintf(hand,"// BE SURE TO SET the type and location of your romset\n");
  fprintf(hand,"// rom set 0 looks like tmnt.d18\n");
  fprintf(hand,"// rom set 1 looks like 963-r23\n");
  fprintf(hand,"// rom set 2 looks like 963-u23 (2-player version)\n");
  fprintf(hand,"// rom set 3 looks like 963-x23 (japan)\n");
  fprintf(hand,"// rom set 4 looks like 963-123 (japan 2-player version)\n");
  fprintf(hand,"rom_set %d\n",rom_set);
  fprintf(hand,"rom_path:%s\n",rom_path);

  fprintf(hand,"// Below is the setup for the screen mode\n\n");
  fprintf(hand,"// Some good screen modes are:\n");
  fprintf(hand,"// GFX_VESA2L (5) 320 240 16-bit, if you have VESA 2.0 or Scitech DD\n");
  fprintf(hand,"// GFX_VESA1 (3) 640 480 16-bit, if you want the color, but only have VESA 1\n");
  fprintf(hand,"// GFX_VGA  (1) 320 200 8-bit, which is stretched but a bit faster\n\n");

  fprintf(hand,"// gr.driver can be:\n\n");
  for (i=0;i<20;i+=2)
  {
    fprintf(hand,"// %2d %-16.16s  %2d %-16.16s\n",
     i,driver_name[i+1],
     i+1,driver_name[i+2]);
  }
  fprintf(hand,"\n");

  fprintf(hand,"gr.driver %d\n",gr.driver);
  fprintf(hand,"gr.x %d\n",gr.x);
  fprintf(hand,"gr.y %d\n",gr.y);
  fprintf(hand,"// gr.depth (color depth) can be 8 15 or 16: (24 or 32 runs slower)\n");
  fprintf(hand,"gr.depth %d\n",gr.depth);
  fprintf(hand,"layer 0x%.2x\n",layer);
  fprintf(hand,"wait_vbl %d\n",wait_vbl);
  fprintf(hand,"auto_frameskip %d\n",auto_frameskip);
  fprintf(hand,"virtual_fps %d\n",virtual_fps);
  fprintf(hand,"// If auto_frameskip is 1, virtual_fps is used,\n");
  fprintf(hand,"// else frame_skip is used\n");
  fprintf(hand,"frame_skip %d\n",frame_skip);

  fprintf(hand,"\n");

  fprintf(hand,"// NB - If you change dip_swicthes, you'll need to reboot the arcade machine.\n");
  fprintf(hand,"// coins per credit (00 - 1 coin/credit)\n");
  fprintf(hand,"dip_switch1 0x%.2x\n",dip[1]);
  fprintf(hand,"// difficulty (easiest 82  a2 c2  e2 hardest) ('2' part means 3 lives per credit) \n");
  fprintf(hand,"dip_switch2 0x%.2x\n",dip[2]);
  fprintf(hand,"// 0x04 = test mode\n");
  fprintf(hand,"dip_switch3 0x%.2x\n",dip[3]);
  fprintf(hand,"\n");
  fprintf(hand,
  "// Fit screen scales the %dx%d to whatever res you are using\n",
  ARCADE_W,ARCADE_H);
  fprintf(hand,"fit_screen %d\n",fit_screen);
  fprintf(hand,"gamma_correct %d\n",gamma_correct);
  fprintf(hand,"gamma_value %.2f\n",gamma_value);
  fprintf(hand,"trippy_sprites %d\n",trippy_sprites);
  fprintf(hand,"// The arcade machine flickers to create transparent effects\n");
  fprintf(hand,"stop_flicker %d\n",stop_flicker);

  fprintf(hand,"\n\n// -------- SOUND CONFIG --------\n");


  fprintf(hand,"// Note - Sound emulation is very slow - turning it off can double the frame rate!\n\n");
  fprintf(hand,"// Set want_sound 0 to turn off sound\n\n");
  fprintf(hand,"// 'using_sound_card' is the sound driver to use\n");
  fprintf(hand,"// Allegro values are:"
  "%d:autodetect "
  "%d:sbpro "
  "%d:sb "
  "%d:sb10 "
  "%d:sb15 "
  "%d:sb20 "
  "%d:SB16-bit "
  "%d:GUS "
  "%d:ESS audiodrive \n",
      DIGI_AUTODETECT,
      DIGI_SBPRO     ,
      DIGI_SB        ,
      DIGI_SB10      ,
      DIGI_SB15      ,
      DIGI_SB20      ,
      DIGI_SB16      ,
      DIGI_GUS       ,
      DIGI_AUDIODRIVE);
  fprintf(hand,"// *** using_digi_card 100 is a special faster driver SBSND\n");
  fprintf(hand,"// which is for SoundBlasters. Try it.\n");
  fprintf(hand,"want_sound %d\n",want_sound);
  fprintf(hand,"using_digi_card %d\n",using_digi_card);
  fprintf(hand,"stereo16bit %d\n",stereo16bit);
  fprintf(hand,"dodgy_drums %d\n",dodgy_drums);
  fprintf(hand,"waveout_size %d\n",waveout_size);
  fprintf(hand,"// Change the quality of the sound (8000=terrible, 22050=good, but very slow)\n");
  fprintf(hand,"waveout_rate %d\n",waveout_rate);
  fprintf(hand,"ym2151_amp %d\n",ym2151_amp);
  fprintf(hand,"drum_volume %d\n",drum_volume);
  fprintf(hand,"alter_pitch %.2f\n",alter_pitch);
  fprintf(hand,"timerb_speed %d\n",timerb_speed);
  fprintf(hand,"\n");
  fprintf(hand,"// Force DSP version is for Allegro's Sound (try 0x300 or 0x201)\n");
  fprintf(hand,"force_dsp_version 0x%x\n",force_dsp_version);
  fprintf(hand,"// The following info is for SBSND only (driver 100)\n");
  fprintf(hand,"get_from_blaster %d\n",get_from_blaster);
  fprintf(hand,"sbsnd_base %x\n",sbsnd_base);
  fprintf(hand,"irq_num %d\n",irq_num);
  fprintf(hand,"dma_chan_8 %d\n",dma_chan_8);
  fprintf(hand,"dma_chan_16 %d\n",dma_chan_16);
  fprintf(hand,"\n\n");

  fprintf(hand,"show_sound_info %d\n",show_sound_info);
  fprintf(hand,"display_scodes %d\n",display_scodes);

  {
    int i;
    fprintf (hand,"scode_remapped\n");
    for (i=0;i<256;i++)
    {
      if (scode_remap[i]!=i)
      {
        fprintf (hand,"scode_remap 0x%.2x 0x%.2x\n",i,scode_remap[i]);
      }
    }
  }

  fprintf(hand,"\n// -----------------------------------------------------\n");

  fprintf(hand,"// Use F12 to toggle emulator_keys_off\n");
  fprintf(hand,"// If emulator_keys are off, only F12 and Escape will work\n");
  fprintf(hand,"// This leaves the whole keyboard free for controllers\n");
  fprintf(hand,"// If emulator_keys_off is -1, only F12 will work!\n");
  fprintf(hand,"// (You might want this so you don't quit by accident.)\n");
  fprintf(hand,"emulator_keys_off %d\n",emulator_keys_off);
  fprintf(hand,"show_fps %d\n",show_fps);

  fprintf(hand,"\n");

  fprintf(hand,"// NOTE - JOY_TYPE_STANDARD (%d) should be okay for any jostick\n",JOY_TYPE_STANDARD);
  fprintf(hand,"set_joy_type %d\n",set_joy_type);

  fprintf(hand,"// The next line says which parallel ports DTMNT should read for SNES joypads\n");
  fprintf(hand,"// 0 (default) means don't read (no joypads)\n");
  fprintf(hand,"// 1 means read LPT1 (2 pads), 2 means read LPT2 (2 pads), and 3 means read BOTH (4 pads total)\n\n");
  fprintf(hand,"lpt_read %d\n\n",lpt_read);

  fprintf(hand,"\n//I quote from Allegro!:
//   %d JOY_TYPE_AUTODETECT
//      Attempts to autodetect your joystick hardware. It isn't possible to 
//      reliably distinguish between all the possible input setups, so this 
//      routine can only ever choose the standard joystick, Sidewider, or 
//      GamePad Pro drivers, but it will use information from the 
//      configuration file if one is available (this can be created using the 
//      setup utility or by calling the save_joystick_data() function), so you 
//      can always use JOY_TYPE_AUTODETECT in your code and then select the 
//      exact hardware type from the setup program.
//
//   %d JOY_TYPE_NONE
//      Dummy driver for machines without any joystick.
//
//   %d JOY_TYPE_STANDARD
//      A normal two button stick.
//
//   %d JOY_TYPE_2PADS
//      Dual joystick mode (two sticks, each with two buttons).
//
//   %d JOY_TYPE_4BUTTON
//      Enable the extra buttons on a 4-button joystick.
//
//   %d JOY_TYPE_6BUTTON
//      Enable the extra buttons on a 6-button joystick.
//
//   %d JOY_TYPE_8BUTTON
//      Enable the extra buttons on an 8-button joystick.
//
//   %d JOY_TYPE_FSPRO
//      CH Flightstick Pro or compatible stick, which provides four buttons, 
//      an analogue throttle control, and a 4-direction coolie hat.
//
//   %d JOY_TYPE_WINGEX
//      A Logitech Wingman Extreme, which should also work with any 
//      Thrustmaster Mk.I compatible joystick. It provides support for four 
//      buttons and a coolie hat. This also works with the Wingman Warrior, if 
//      you plug in the 15 pin plug (remember to unplug the 9-pin plug!) and 
//      set the tiny switch in front to the \"H\" position (you will not be able 
//      to use the throttle or the spinner though).
//
//   %d JOY_TYPE_SIDEWINDER
//      The Microsoft Sidewinder digital pad (supports up to four 
//      controllers, each with ten buttons and a digital direction control).
//
//   %d JOY_TYPE_GAMEPAD_PRO
//      The Gravis GamePad Pro (supports up to two controllers, each with ten 
//      buttons and a digital direction control).
\n//Phew!\n\n",
   JOY_TYPE_AUTODETECT,
   JOY_TYPE_NONE,
   JOY_TYPE_STANDARD,
   JOY_TYPE_2PADS,
   JOY_TYPE_4BUTTON,
   JOY_TYPE_6BUTTON,
   JOY_TYPE_8BUTTON,
   JOY_TYPE_FSPRO,
   JOY_TYPE_WINGEX,
   JOY_TYPE_SIDEWINDER,
   JOY_TYPE_GAMEPAD_PRO
  );

  fprintf(hand,"\n");
  fprintf(hand,"//Next is a huge chunk of codes: they are the input codes\n");
  fprintf(hand,"//for the system (coins/service) 0 thru raph (4)\n");
  fprintf(hand,"//The order is left right up down jump attack button3 button4\n");
  fprintf(hand,"//(It's a lot easier to just hit R in the emu!)\n\n");

  {
    int p,k;
    for (p=0;p<5;p++)
    {
      for (k=0;k<8;k++)
      {
        fprintf (hand,
          "input_type[%d][%d] 0x%.4x\n",p,k,input_type[p][k]);
      }
    }
  }
  fprintf(hand,"//End of config!\n");

  fclose(hand);

  // delete the sound code remaps
  if (scode_remap!=NULL)
    { tree(scode_remap); scode_remap=NULL; }

  return 0;
}
