// DTMNT emu - initialisation stuff
// (for our end - DOS)
#include <stdio.h>
#include <stdlib.h>
#include <allegro.h>
#include "dtmnt.h"

volatile long long tenths=0;
volatile long long hz60=0;
volatile unsigned char ym2151_timers=0;
PALETTE old_pal;

// need more timer accuracy for the sound interrupts

void called_aimed_for_framerate_each_sec()
{
  if (run_timer) hz60++;
}
END_OF_FUNCTION(called_aimed_for_framerate_each_sec);

void called_ten_times_each_sec()
{
  if (run_timer) tenths++;
  if (message_count>0) message_count--;
}
END_OF_FUNCTION(called_ten_times_each_sec);

int various_init()
{
  {
    extern int record_playback;
    if (record_playback>0)
    {
      if (save_input_init("movie.tmt")!=0)
      {
        record_playback=0;
        printf ("\n---*- FAILED TO SAVE INPUT to movie.tmt -*----\n\n");
      }
      else
        printf ("\n----- SAVING INPUT to movie.tmt ------\n\n");
    }
    else if (record_playback<0)
    {
      if (load_input_init("movie.tmt")!=0)
      {
        record_playback=0;
        printf ("\n---*- FAILED TO LOAD INPUT from movie.tmt -*----\n\n");
      }
      else
        printf ("\n----- LOADING INPUT from movie.tmt ------\n\n");
    }
  }


  if (allegro_init()!=0)
    printf ("Error installing allegro!\n");
  printf ("Made using %s\n",allegro_id);

  joy_type=set_joy_type;
  if (initialise_joystick()!=0)
    printf ("Could not initialise joystick.\n");
  if (install_mouse()<1)
    printf ("Error installing mouse!\n");
  if (install_timer()!=0)
    printf ("Error installing timer!\n");
  if (install_keyboard()!=0)
    printf ("Error installing keyboard!\n");

  // Lock the timer affected stuff in memory
  LOCK_VARIABLE(run_timer);
  LOCK_VARIABLE(hz60);
  LOCK_VARIABLE(message_count);
  LOCK_VARIABLE(tenths);
  LOCK_VARIABLE(ym2151_timers);
  LOCK_FUNCTION(called_aimed_for_framerate_each_sec);
  LOCK_FUNCTION(called_ten_times_each_sec);

  if (install_int_ex(called_ten_times_each_sec,BPS_TO_TIMER(10))!=0)
    printf ("Error installing 10 bps timer!\n");


  sound_init();

  if (virtual_fps>200) virtual_fps=200;
  if (virtual_fps<1)   virtual_fps=1;

  if (
       install_int_ex
       (
         called_aimed_for_framerate_each_sec,
         BPS_TO_TIMER(virtual_fps)
       )
       !=0
     )
  {
    printf ("Error installing %d fps timer!\n",virtual_fps);
    auto_frameskip=0;
    printf ("(You can't use auto_frameskip.)\n");
  }

  // Try the 'gr' screen mode
  actual=gr;
  get_palette(old_pal);

  printf ("Going to try this mode: %s %dx%d %d-bit\n",
    driver_name[(actual.driver+1)&31],actual.x,actual.y,actual.depth);
  printf ("\n\nPress a key...\n");
  while (!keypressed())
    ;
  if ((readkey()&255)==27) { return 1; }


  set_color_depth(actual.depth);
  if (set_gfx_mode(actual.driver,actual.x,actual.y,0,0)!=0)
  {
    int key=0;
    printf ("Couldn't do that screen mode - sorry... Press a key to default to 320x200.\n");
    printf ("(Or press F8 to carry on - won't usually help.)\n");
    while (!keypressed())
      ;
    key=readkey();
    if ((key&255)==27) { return 1; }
    if ((key>>8)!=KEY_F8)
    {
      actual.driver=GFX_VGA;
      actual.x=320; actual.y=200;
      actual.depth=8;
      set_color_depth(actual.depth);
      set_gfx_mode(actual.driver,actual.x,actual.y,0,0);
    }
  }

  if (actual.depth==8)
  {
    // make a nice palette
    generate_332_palette(game_pal);
    game_pal[0].r=0;
    game_pal[0].g=0;
    game_pal[0].b=0;
    rgb_map=tralloc(32*32*32);
    if (rgb_map!=NULL)
      create_rgb_table(rgb_map,game_pal,NULL);
    set_palette(game_pal);
  }

  return 0;
}

int various_exit()
{
  sound_exit();
  set_palette(old_pal);
  allegro_exit();
  if (rgb_map!=NULL) tree(rgb_map);

  saveload_input_exit();

  return 0;
}

int various_info(int total_kilo_cycles)
{
  printf ("\nUsed this mode: %s %dx%d %d-bit\n",
    driver_name[(actual.driver+1)&31],actual.x,actual.y,actual.depth);
  {
    int virtual_tenths=total_kilo_cycles/(SPEED_OF_68K/10000);
    long long real_tenths=tenths;
    int percent=0;
    printf ("Real time %lld.%.2lld seconds.\n",real_tenths/10,real_tenths%10);
    printf ("(Virtual time %d.%d seconds)\n\n",virtual_tenths/10,virtual_tenths%10);
    {
      int virt_perc=(virtual_fps*1000)/60;
      printf ("virtual_fps was %d (%d.%d%% of arcade's speed)\n",
        virtual_fps,virt_perc/10,virt_perc%10);
    }
    if (real_tenths)
    {
      percent=(virtual_tenths*1000)/real_tenths;
      printf ("emu ran at %d.%d%% of arcade's speed.\n",percent/10,percent%10);
    }
    printf ("\n");
  }
  return 0;
}

