// DTMNT emu - main module

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <allegro.h>
#include "dtmnt.h"

FILE *debug_log=NULL;

// This is the mode we asked for
struct graph_mode gr={GFX_MODEX,320,240,8};
struct graph_mode actual={GFX_VGA,320,200,8};

static char config_file[128]="dtmnt.ini";

static unsigned char temp[1024]="";
char rom_path[1024]="roms\\";
int layer=15;
PALETTE game_pal;
int frame_skip=2,auto_frameskip=1,virtual_fps=60;
int wait_vbl=0;
int fit_screen=0;
int pause=0,run_timer=0;
int emulator_keys_off=0;
int show_fps=0;
int show_sound_info=0;
char message[256]=""; volatile int message_count=0;
struct mescol mescol={0,0,0},mes_white={255,240,230},
  mes_red={255,180,150},mes_blue={140,200,255};

static int take_screenshot();

static long long total_allocated_memory=0;
void *tralloc(int size)
{
  void *mem;
  mem=malloc(size);
  if (mem!=NULL) total_allocated_memory+=size;
  return mem;
}

int tree(void *mem)
{
  if (mem!=NULL) free(mem);
  return 0;
}

#define FPS_ACC (32)

int main(int argc,char **argv)
{
  int done=0;
  BITMAP *scrbuf;
  int total_kilo_cycles=0;
  long long frame_count=0;
  static int fps=0,fps_tenths_history[FPS_ACC]={0};

//  debug_log=fopen("debug.log","wt");

  printf ("======= Dave's TMNT emulator =======\n");
  printf ("Version %s - compiled %s\n\n",VER,__DATE__);

  rgb_map=NULL;

  if (load_config(config_file)!=0)
  {
    printf (
      "(Create a default config file DTMNT.INI\n"
      " - please edit this to suit your computer...)\n\n");
    unload_config(config_file);

    printf ("(To reset to default settings, simply delete DTMNT.INI and rerun program.)\n");

    printf ("\n\nPress a key to continue with default setup, or escape to exit.\n");
    while (!kbhit())
      ;
    if ((getkey()&255)==27) { return 0; }
    // HAVE TO RELOAD config file (frees/reallocs some memory!)
    load_config(config_file);
  }

  command_line_options(argc,argv);

  run_timer=0; hz60=0; // don't start our timer running yet
  if (machine_init()!=0) exit(1);

  if (various_init()!=0) { various_exit(); exit(1); }

  scrbuf=create_bitmap(ARCADE_W+16,ARCADE_H+16);
  // 8 pixels either side, so we don't have to worry about tile clipping
  if (scrbuf==NULL)
    { various_exit(); printf ("Couldn't make scrbuf\n"); exit(1); }
  clear(scrbuf);

  if (emulator_keys_off==-1)
  {
    sprintf(message,"Emulator keys ALL off! - F12 to unlock");
    mescol=mes_red; message_count=120;
  }
  else
  {
    sprintf(message,"Dave's TMNT emu v%s",VER);
    mescol=(struct mescol){128,255,120}; message_count=60;
  }

  frame_count=0;

  // Even if we don't execute any frames, we still need
  // to see were the sprites are...
  graphics_info();

  run_timer=1; hz60=0; // start timing

  while (!done)
  {
    int advance=0,screen_shot=0,changed_layer=0,changed_frame_skip=0;
    static int enter_scode=0;
    static unsigned char sctext[16]="";
    static long long last_hz60=0;
    static int show_fps_going=0;
    static int save_slot=0;

    SBK

    if (enter_scode)
    {
      sprintf(message,"Sound code 0x%s_",sctext);
      mescol=mes_blue;
      message_count=-1;
    }

    SBK
    while (keypressed())
    {
      int key=readkey(),scan;
      scan=(key>>8)&0xff;
      key&=0xff;

// moved this up - I guess that's why F12 didn't work!
      if (scan==KEY_F12)
      {
        emulator_keys_off=!emulator_keys_off;
        sprintf(message,"(F12) Emulator keys %s",emulator_keys_off?"off":"on");
        mescol=emulator_keys_off ? mes_red:mes_blue; message_count=30;
      }

      if (emulator_keys_off==-1) continue;

      if (scan==KEY_ESC) done=1;

      if (emulator_keys_off) continue;
      
      if (enter_scode)
      {
        // Enter a sound code for the z80
        if ((scan==KEY_ENTER)||(enter_scode>8))
        {
          int code=0;
          code=strtol(sctext,NULL,16);

          sound_effect[sound_effect_count]=code;
          if (sound_effect_count<(SOUND_EFFECT_LIMIT-1))
            sound_effect_count++;

          sprintf(message,"Sending :  0x%.2x",code);
          mescol=mes_blue;
          message_count=15;

          enter_scode=0;
        }
        else
        {
          sctext[enter_scode-1]=key;
          sctext[enter_scode]=0;
          enter_scode++;
        }
        // Don't count the keypresses for anything else
        continue;
      }
      if (scan==KEY_E) { enter_scode=1; sctext[0]=0; }

    SBK

      if (scan==KEY_W)
      {
        display_scodes=(display_scodes+1)%3;
        if (display_scodes==0) sprintf(message,"Sound Codes: hidden");
        if (display_scodes==1) sprintf(message,"Sound Codes: Some displayed");
        if (display_scodes==2) sprintf(message,"Sound Codes:  All displayed");
        mescol=mes_blue;
        message_count=20;
      }
      
           if (scan==KEY_1) { layer^=1; changed_layer=1; }
      else if (scan==KEY_2) { layer^=2; changed_layer=1; }
      else if (scan==KEY_3) { layer^=4; changed_layer=1; }
      else if (scan==KEY_4) { layer^=8; changed_layer=1; }
      else if (scan==KEY_9) { layer=15; changed_layer=1; }
      else if (scan==KEY_0) { layer= 8; changed_layer=1; }
      else if (scan==KEY_F10)
      {
        do_int5=0;
        sprintf(message,"do_int5 %d",do_int5);
        mescol=mes_white; message_count=15;
      }
      else if (scan==KEY_F11)
      {
        do_int5=1;
        sprintf(message,"do_int5 %d",do_int5);
        mescol=mes_white; message_count=15;
      }

      else if (scan==KEY_F9) screen_shot=1;
      else if (scan==KEY_P) pause=!pause;
      else if (scan==KEY_PAUSE) pause=!pause;
      else if (scan==KEY_A) advance=1;
      else if (scan==KEY_O)
      {
        static char tt[4][16]={"none","wave","boox","boox and wave"};
        show_sound_info++;
        show_sound_info&=3;
        sprintf(message,"Sound info : %s",tt[show_sound_info]);
        mescol=mes_white; message_count=15;
      }
      else if (scan==KEY_T)
      {
        trippy_sprites=(((trippy_sprites+2)%2)-1);
        sprintf(message,"Trippy sprites : %d",trippy_sprites);
        mescol=mes_white; message_count=15;
      }
      else if (scan==KEY_S)
      {
        auto_frameskip=!auto_frameskip;
        sprintf(message,"Auto frameskip : %s",auto_frameskip ? "ON":"off");
        mescol=mes_white; message_count=15;
      }
      else if (scan==KEY_F)
      {
        fit_screen=(fit_screen+1)%3;
        sprintf(message,"Fit screen : %d",fit_screen);
        mescol=mes_white; message_count=15;
      }
      else if (scan==KEY_I)
      {
        show_fps=!show_fps;
        sprintf(message,"Show f.p.s %s",show_fps ? "on":"off");
        mescol=mes_white; message_count=15;
      }
      else if (scan==KEY_EQUALS) {frame_skip++; changed_frame_skip=1;}
      else if (scan==KEY_MINUS) {frame_skip--; changed_frame_skip=1;}

      else if (scan==KEY_V)
      {
        wait_vbl=!wait_vbl;
        sprintf(message,"Wait vbl : %s",wait_vbl ? "ON":"off");
        mescol=mes_white; message_count=15;
      }
      else if (scan==KEY_R)
      {
        redefine_input();
        sprintf(message,"Input redefining done");
        mescol=mes_white; message_count=10;
      }
      else if (scan==KEY_F8)
      {
        int i=0;
        save_slot++; save_slot&=3;
        sprintf(message,"Save slot ");
        for (i=0;i<4;i++)
        {
          sprintf(temp,"%c%c%c  ",
            i==save_slot ? '[':' ',
            i==save_slot ? 'A'+i : 'A'+i,
            i==save_slot ? ']':' ');
          strcat(message,temp);
        }
        mescol=mes_blue; message_count=10;
      }
      else if (scan==KEY_F7)
      {
        if (key_shifts&KB_SHIFT_FLAG)
        {
          if (record_playback<0)
          {
            record_playback=0;
            sprintf(message,"Movie playback stopped");
          }
          else
          {
            saveload_input_exit();
            sprintf (temp,"movie%c.tmt",'A'+save_slot);
            if (load_input_init(temp)!=0)
              sprintf(message,"Error playing Movie from Slot %c",'A'+save_slot);
            else
            {
              record_playback=-1;
              sprintf(message,"Playing Movie from Slot %c",'A'+save_slot);
            }
          }
        }
        else
        {
          sprintf (temp,"tmnt%c.sav",'A'+save_slot);
          if (machine_load(temp)==0)
            sprintf(message,"Loaded from slot %c",'A'+save_slot);
          else
            sprintf(message,"Error loading from %s",temp);
        }
        mescol=mes_blue; message_count=10;
      }
      else if (scan==KEY_F5)
      {
        if (key_shifts&KB_SHIFT_FLAG)
        {
          if (record_playback>0)
          {
            record_playback=0;
            sprintf(message,"Movie recording stopped");
          }
          else
          {
            saveload_input_exit();
            sprintf (temp,"movie%c.tmt",'A'+save_slot);
            if (save_input_init(temp)!=0)
              sprintf(message,"Error saving Movie to Slot %c",'A'+save_slot);
            else
            {
              record_playback=1;
              sprintf(message,"Saving Movie to Slot %c",'A'+save_slot);
            }
          }
        }
        else
        {
          sprintf (temp,"tmnt%c.sav",'A'+save_slot);
          if (machine_save(temp)==0)
            sprintf(message,"Saved to slot %c",'A'+save_slot);
          else
            sprintf(message,"Error saving to %s",temp);
        }
        mescol=mes_red; message_count=10;
      }
    }
    SBK


    if (!show_fps)
    {
      show_fps_going=0;
    }
    else
    {
      // Figure out the fps count:
      // We have a record of the number of tenths FPS_ACC frames ago
      // nb FPS_ACC must be a power of 2
      long long old_tenths=fps_tenths_history[frame_count&(FPS_ACC-1)];
      long long new_tenths=tenths,diff_tenths;

      fps_tenths_history[frame_count&(FPS_ACC-1)]=new_tenths;

      diff_tenths=new_tenths-old_tenths;

      // So fps is FPS_ACC/(tenths/10) or (FPS_ACC*10)/(diff_tenths)
      fps=0;
      if (diff_tenths)
        fps=((FPS_ACC)*100)/diff_tenths; // times 10
      if (show_fps_going<FPS_ACC)
        show_fps_going++;
    }
    frame_count++;

    SBK

    if (changed_layer)
    {
      sprintf(message,"Layer : (%c%c%c%c) 0x%.2x",
        (layer&1)?'T':' ',
        (layer&2)?'F':' ',
        (layer&4)?'B':' ',
        (layer&8)?'S':' ',
        layer
      );
      mescol=mes_white; message_count=15;
    }

    if (frame_skip<1) frame_skip=1;

    if (changed_frame_skip)
    {
      sprintf(message,"Frame skip : %2d ",frame_skip);
      mescol=mes_white; message_count=15;
    }

    // Update 68000
    if (pause)
    {
      if (advance)
      {
        machine_do_one_frame();
      }
      graphics_info();
      last_hz60=hz60; // so it doesn't jump when we unpause!
    }
    else
    {
      int i;
      if (auto_frameskip)
      {
        int frames_gone=hz60-last_hz60;
        last_hz60=hz60;
        if (frames_gone>20) frames_gone=20;
        for (i=0;i<frames_gone;i++)
        {
          machine_do_one_frame();
          graphics_info();
        }
      }
      else
      {
        last_hz60=hz60;
        for (i=0;i<frame_skip;i++)
        {
          machine_do_one_frame();
          graphics_info();
        }
      }
    }

    SBK
    draw_graphics(scrbuf);
    SBK
    {
      extern int draw_waveout(BITMAP *bmp);
      if (show_sound_info&1) draw_waveout(scrbuf);
    }
    SBK

    {
      static int last_fit_screen=0;
      int borderx=(SCREEN_W-ARCADE_W)/2;
      int bordery=(SCREEN_H-ARCADE_H)/2;

      if (screen_shot)
      {run_timer=0; take_screenshot(); run_timer=1; }

      if (wait_vbl) vsync();

      if (last_fit_screen && (!fit_screen)) clear(screen);
      last_fit_screen=fit_screen;


    SBK

      if (message_count)
      {
        int meslen=0;
        meslen=text_length(font,message);
        rectfill(scrbuf,8+1,scrbuf->h-24-9,8+2+meslen+4,scrbuf->h-24-8+8+2,
          makecol(mescol.r/2,mescol.g/2,mescol.b/2));
        text_mode(-1);
        textout(scrbuf,font,message,8+4,
          scrbuf->h-22-8,makecol(mescol.r*3/4,mescol.g*3/4,mescol.b*3/4));
        text_mode(-1);
        textout(scrbuf,font,message,8+3,
          scrbuf->h-23-8,makecol(mescol.r,mescol.g,mescol.b));
      }
      if (pause)
      {
        text_mode(0);
        textout(scrbuf,font,"Paused",8+4,
          scrbuf->h-40-8,makecol(128,128,128));
        text_mode(-1);
        textout(scrbuf,font,"Paused",8+3,
          scrbuf->h-41-8,makecol(255,255,255));
      }


    SBK

      if (show_fps)
      {
        if (show_fps_going<FPS_ACC)
          sprintf (temp,"    fps (wait)");
        else
          sprintf (temp,"%3d fps",fps/10);

        text_mode(0);
        textout(scrbuf,font,temp,8+4,
          scrbuf->h-32-8,makecol(80,40,100));
        text_mode(-1);
        textout(scrbuf,font,temp,8+3,
          scrbuf->h-33-8,makecol(140,200,250));
      }

    SBK

      if (show_sound_info&2)
      {
        extern unsigned char booreg[16];
        int i;
        sprintf (temp,"boox ");
        for (i=0;i<16;i++)
        {
          sprintf (temp+512,"%2x",i);
          strcat(temp,temp+512);
        }
        text_mode(0);
        textout(scrbuf,font,temp,8+4,
          scrbuf->h-52-18,makecol(80,40,100));
        text_mode(-1);
        textout(scrbuf,font,temp,8+3,
          scrbuf->h-53-18,makecol(240,200,250));

        sprintf (temp,"     ");
        for (i=0;i<16;i++)
        {
          sprintf (temp+512,"%.2x",booreg[i]);
          strcat(temp,temp+512);
        }
        text_mode(0);
        textout(scrbuf,font,temp,8+4,
          scrbuf->h-52-8,makecol(80,40,100));
        text_mode(-1);
        textout(scrbuf,font,temp,8+3,
          scrbuf->h-53-8,makecol(140,200,250));
      }

      {
        int col=0,shade=0;
        temp[0]=' ';
        temp[1]=0;
        if (record_playback>0)
          { temp[0]='*'; col=makecol(255,100,100); shade=makecol(100,0,0); }
        if (record_playback<0)
          { temp[0]='R'; col=makecol(100,255,100); shade=makecol(0,100,0);}
        
        text_mode(-1);
        textout(scrbuf,font,temp,scrbuf->w-15,
          scrbuf->h-15,shade);
        textout(scrbuf,font,temp,scrbuf->w-16,
          scrbuf->h-16,col);
      }

    SBK

      if (gr.driver!=GFX_TEXT)
      switch(fit_screen)
      {
        // Choice of missing messy tile clipping bits by taking middle region
        case 1:
          stretch_blit(scrbuf,screen,8,8,ARCADE_W,ARCADE_H,0,0,SCREEN_W,SCREEN_H);
          break;
        case 2:
          stretch_blit(scrbuf,screen,0,0,scrbuf->w,scrbuf->h,0,0,SCREEN_W,SCREEN_H);
          break;
        default:
          blit(scrbuf,screen,8,8,borderx,bordery,ARCADE_W,ARCADE_H);
      }
    SBK
    }
  }

  run_timer=0; // freeze our own timer
  destroy_bitmap(scrbuf);

  various_exit();

  total_kilo_cycles=machine_exit();

  various_info(total_kilo_cycles);
  printf("Rendered %lld frames\n",frame_count);

  // It's confusing if the keys don't come on next time?
  emulator_keys_off=0;
  unload_config(config_file);

  printf (
    "\n\n-----------------------------------------\n"
    " Current homepage for DTmnt:\n"
    "         http://cybermillenia.com/dayvee/\n"
    " Old homepage:                           \n"
    "http://www.geocities.com/Paris/Parc/4330/\n"
    "Emulator written by Dave:\n"
    "dayvee@rocketmail.com\n"
    "-----------------------------------------\n\n"
  );

  if (debug_log!=NULL) fclose(debug_log);

  return 0;
}

int take_screenshot()
{
  int foundone=0;
  static int screenshot_count=0;
  // Take a screen shot


  foundone=0;
  while ((!foundone)&&(screenshot_count<9999))
  {
    FILE *hand;
    sprintf (temp,"tmnt%.4d.pcx",screenshot_count);
    hand=fopen(temp,"rb");
    if (hand==NULL) foundone=1;
    else
    {
      fclose(hand);
      screenshot_count++;
    }
  }

  if (foundone)
  {
    BITMAP *renderbmp=NULL;
    BITMAP *savebmp=NULL;
    // Make a high quality version
    renderbmp=create_bitmap_ex(16,ARCADE_W+16,ARCADE_H+16);
    if (renderbmp==NULL) return 1;
    draw_graphics(renderbmp);

    savebmp=renderbmp;
    if (fit_screen!=2)
    {
      // Don't take the messy border
      savebmp=create_sub_bitmap(renderbmp,8,8,ARCADE_W,ARCADE_H);
    }

    save_pcx(temp,savebmp,game_pal);
    // Changed because Allegro crashed

    if (fit_screen!=2)
      destroy_bitmap(savebmp);

    sprintf(message,"Saved rendering to %s",temp); mescol=mes_white; message_count=15;

    destroy_bitmap(renderbmp);

    screenshot_count++;
  }

  return 0;
}
