// DTMNT emu
// Rom loading

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <allegro.h>
#include "mz80.h"
// StarScream specific start!
#include "starcpu.h"
// StarScream specific end!
#include "dtmnt.h"

static unsigned char temp[512];

unsigned char *z80_mem;
unsigned char *sprite4=NULL,*sprite6=NULL,*tile27=NULL;
unsigned char *rom68k=NULL; // 000000 to 05ffff
unsigned char *rammem=NULL; // 060000 to 063fff
unsigned char *colmem=NULL; // 080000 to 080fff
unsigned char *coomem=NULL; // 0c0000 to 0c00ff
unsigned char *vidmem=NULL; // 100000 to 107fff
unsigned char *sprmem=NULL; // 140400 to 1407ff
unsigned char do_int5=0;

static int cycles=0,mega_cycles=0;
static int extra_cycles=0; //extra cycles we did last time
static int last_sprites_behind=0;
int reboot=0,save_state=1;

// Return -1 if no file could be opened, and 1 if info is missing
char autosave_name[128]="tmntmem.sav";

// StarScream specific start!
int machine_save_to(FILE *hand)
{
  static char id[64]="DTmnt Save --- v" VER;
  struct S68000CONTEXT cont;
  static struct mz80context z80={0};
  extern int pause;

  // Save current state (but don't dealloc anything)

  fwrite(id,1,64,hand);
  s68000GetContext(&cont);
  fwrite(cont.dreg,4,8,hand);
  fwrite(cont.areg,4,8,hand);
  fwrite(&cont.asp,4,1,hand);
  fwrite(&cont.pc,4,1,hand);
  fwrite(&cont.sr,2,1,hand);
  fwrite(&cont.stopped,1,1,hand);
  fwrite(&do_int5,1,1,hand);
  fwrite(&extra_cycles,4,1,hand);
  fwrite(&pause,4,1,hand);
  fwrite(&dip[1],1,1,hand);
  fwrite(&dip[2],1,1,hand);
  fwrite(&dip[3],1,1,hand);
  fseek(hand,0x69,SEEK_CUR);

  fwrite(rammem,1,0x4000,hand);
  fwrite(colmem,1,0x1000,hand);
  fwrite(coomem,1,0x0100,hand);
  fwrite(vidmem,1,0x8000,hand);
  fwrite(sprmem,1,0x0400,hand);
  fwrite(&last_sprites_behind,4,1,hand);
  fseek(hand,0xafc,SEEK_CUR);

  mz80GetContext(&z80);

  fwrite(&z80.z80clockticks,4,1,hand);
  fwrite(&z80.z80inInterrupt,4,1,hand);
  fwrite(&z80.z80interruptMode,4,1,hand);
  fwrite(&z80.z80interruptState,4,1,hand);
  fwrite(&z80.z80halted,4,1,hand);
  fwrite(&z80.z80af,2,1,hand);
  fwrite(&z80.z80bc,2,1,hand);
  fwrite(&z80.z80de,2,1,hand);
  fwrite(&z80.z80hl,2,1,hand);
  fwrite(&z80.z80afprime,2,1,hand);
  fwrite(&z80.z80bcprime,2,1,hand);
  fwrite(&z80.z80deprime,2,1,hand);
  fwrite(&z80.z80hlprime,2,1,hand);
  fwrite(&z80.z80ix,2,1,hand);
  fwrite(&z80.z80iy,2,1,hand);
  fwrite(&z80.z80sp,2,1,hand);
  fwrite(&z80.z80pc,2,1,hand);
  fwrite(&z80.z80nmiAddr,2,1,hand);
  fwrite(&z80.z80intAddr,2,1,hand);
  fwrite(&z80.z80i,1,1,hand);
  fwrite(&z80.z80r,1,1,hand);
  fseek(hand,0x4e,SEEK_CUR);

  if (z80_mem!=NULL)
    fwrite(z80_mem+0x8000,1,0x8000,hand);

  return 0;
}

int machine_load_from(FILE *hand)
{
  struct S68000CONTEXT cont;
  static struct mz80context z80={0};
  static char id[64]="";
  extern int pause;
  // Load state

  if (fread(id,1,64,hand)          !=64)  return 1;
  s68000GetContext(&cont);
  if (fread(cont.dreg,4,8,hand)    !=8)   return 1;
  if (fread(cont.areg,4,8,hand)    !=8)   return 1;
  if (fread(&cont.asp,4,1,hand)    !=1)   return 1;
  if (fread(&cont.pc,4,1,hand)     !=1)   return 1;
  if (fread(&cont.sr,2,1,hand)     !=1)   return 1;
  if (fread(&cont.stopped,1,1,hand)!=1)   return 1;
  if (fread(&do_int5,1,1,hand)     !=1)   return 1;
  if (fread(&extra_cycles,4,1,hand)!=1)   return 1;
  if (fread(&pause,4,1,hand)       !=1)   return 1;
  if (fread(&dip[1],1,1,hand)!=1) return 1;
  if (fread(&dip[2],1,1,hand)!=1) return 1;
  if (fread(&dip[3],1,1,hand)!=1) return 1;
  fseek(hand,0x69,SEEK_CUR);
  s68000SetContext(&cont);

  if (fread(rammem,1,0x4000,hand) !=0x4000) return 1;
  if (fread(colmem,1,0x1000,hand) !=0x1000) return 1;
  if (fread(coomem,1,0x0100,hand) !=0x0100) return 1;
  if (fread(vidmem,1,0x8000,hand) !=0x8000) return 1;
  if (fread(sprmem,1,0x0400,hand) !=0x0400) return 1;
  if (fread(&last_sprites_behind,4,1,hand)!=1)   return 1;
  fseek(hand,0xafc,SEEK_CUR);

  mz80GetContext(&z80);
  if (fread(&z80.z80clockticks,4,1,hand)    !=1) return 1;
  if (fread(&z80.z80inInterrupt,4,1,hand)   !=1) return 1;
  if (fread(&z80.z80interruptMode,4,1,hand) !=1) return 1;
  if (fread(&z80.z80interruptState,4,1,hand)!=1) return 1;
  if (fread(&z80.z80halted,4,1,hand)        !=1) return 1;
  if (fread(&z80.z80af,2,1,hand)            !=1) return 1;
  if (fread(&z80.z80bc,2,1,hand)            !=1) return 1;
  if (fread(&z80.z80de,2,1,hand)            !=1) return 1;
  if (fread(&z80.z80hl,2,1,hand)            !=1) return 1;
  if (fread(&z80.z80afprime,2,1,hand)       !=1) return 1;
  if (fread(&z80.z80bcprime,2,1,hand)       !=1) return 1;
  if (fread(&z80.z80deprime,2,1,hand)       !=1) return 1;
  if (fread(&z80.z80hlprime,2,1,hand)       !=1) return 1;
  if (fread(&z80.z80ix,2,1,hand)            !=1) return 1;
  if (fread(&z80.z80iy,2,1,hand)            !=1) return 1;
  if (fread(&z80.z80sp,2,1,hand)            !=1) return 1;
  if (fread(&z80.z80pc,2,1,hand)            !=1) return 1;
  if (fread(&z80.z80nmiAddr,2,1,hand)       !=1) return 1;
  if (fread(&z80.z80intAddr,2,1,hand)       !=1) return 1;
  if (fread(&z80.z80i,1,1,hand)             !=1) return 1;
  if (fread(&z80.z80r,1,1,hand)             !=1) return 1;
  fseek(hand,0x4e,SEEK_CUR);
  if (z80_mem!=NULL)
    fread(z80_mem+0x8000,1,0x8000,hand);
  mz80SetContext(&z80);

  return 0;
}


int machine_save(char *name)
{
  int ret=0;
  FILE *hand=NULL;
  hand=fopen(name,"wb");
  if (hand==NULL) return -1;
  ret=machine_save_to(hand);
  fclose(hand);
  return ret;
}

int machine_load(char *name)
{
  int ret=0;
  FILE *hand=NULL;
  hand=fopen(name,"rb");
  if (hand==NULL) return -1;
  ret=machine_load_from(hand);
  fclose(hand);
  return ret;
}

// StarScream specific end!

// Pass skip as 1 to load normal, 2 to load every byte into 0,2,4
// 3 to words into 01,45,89,cd .. etc (note - 68k order!)
int load_rom(unsigned char *into,char *filename,int skip,int expected_size)
{
  unsigned char *temp=NULL,*t;
  int i=0,len=0;
  FILE *hand=fopen(filename,"rb");
  if (hand==NULL) {printf ("WARNING - Couldn't load %s\n",filename); return 1;}
  fseek(hand,0,SEEK_END);
  len=ftell(hand);
  fseek(hand,0,SEEK_SET);

  if (len!=expected_size)
  {
    printf ("WARNING - rom %s is wrong size: 0x%x/0x%x\n",filename,len,expected_size);
  }

  if (len>expected_size)
  {
    len=expected_size;
    printf ("Only loading 0x%x bytes \n",len);
  }

  temp=tralloc(len);
  if (temp==NULL) { printf ("Couldn't malloc temporary buffer.\n"); return 1;}
  fread(temp,1,len,hand);
  fclose(hand);
  // Now copy into real location

  if (skip==2)
  {
    t=temp;
    for (i=0;i<len;i++)
    {
      *into=*(t++);
      into+=2;
    }
  }
  else if (skip==3)
  {
    t=temp;
    for (i=0;i<(len/2);i++)
    {
      *(into++)=*(t++);
      *(into++)=*(t++);
      into+=2;
    }
  }
  else
  {
    memcpy(into,temp,len);
  }

  tree(temp);
  printf ("> Loaded %s\n",filename); fflush(stdout);
  return 0;
}

// Byteswaps memory
static int byteswap_memory(unsigned char *start,int len)
{
  int i; unsigned char tmp;
  for (i=0;i<len;i+=2)
  {
    tmp=start[i+0];
    start[i+0]=start[i+1];
    start[i+1]=tmp;
  }
  return 0;
}

// Since the tile roms are 4bits per pixel
// I can squeeze 8 pixels in one longword - the only
// trouble is they are the wrong way round if you directly access the ROM!
// Hence - tile_descramble() - it reverses each nibble in 4 bytes
static int tile_descramble(unsigned char *start,int len)
{
  int i;
  register unsigned char t[4];

  for (i=0;i<len;i+=4)
  {
    t[3]=start[i+0];
    t[2]=start[i+1];
    t[1]=start[i+2];
    t[0]=start[i+3];

    t[0]=(t[0]<<4)+(t[0]>>4);
    t[1]=(t[1]<<4)+(t[1]>>4);
    t[2]=(t[2]<<4)+(t[2]>>4);
    t[3]=(t[3]<<4)+(t[3]>>4);

    start[i+0]=t[0];
    start[i+1]=t[1];
    start[i+2]=t[2];
    start[i+3]=t[3];
  }
  return 0;
}

int graphics_info()
{
  extern int we_do_sprites_behind,stop_flicker;
  int sprites_behind=0;

  sprites_behind=(MEMB(coomem,0x00001)&0x04);

  if (stop_flicker)
  {
    if (last_sprites_behind==sprites_behind)
      we_do_sprites_behind=sprites_behind;
    else // it's doing the flicker
      we_do_sprites_behind=1;
  }
  else
    we_do_sprites_behind=sprites_behind;

  last_sprites_behind=sprites_behind;
  work_out_sprites(); // and stash them in a frame record
  return 0;
}


int machine_do_one_frame()
{
  int cycles_we_want_to_do=((SPEED_OF_68K/60)-extra_cycles);
  int cycles_we_did=0;
  extern int record_playback;


  // 68k will read the key in input[5][8], so better set them:

  if (record_playback<0)
  {
    if (load_input()!=0)
      record_playback=0;
  }
  else
  {
    input_update_mouse();
    update_input();
    if (record_playback>0)
    {
      if (save_input()!=0)
        record_playback=0;
    }
  }

  SBK

  // StarScream specific start!

  if (do_int5) s68000interrupt(5);
  s68000context.odometer=0;
  (void)s68000exec(cycles_we_want_to_do);
  cycles_we_did        +=s68000context.odometer;
  cycles_we_want_to_do -=s68000context.odometer;
  extra_cycles=cycles_we_want_to_do;
  // StarScream specific end!

  SBK

  // check the sound stuff (z80/ym2151/our waveout)
  if (do_sound) waveout_stream_update();

  cycles+=cycles_we_did;
  while (cycles>1000000) { cycles-=1000000; mega_cycles++; }


  SBK


  return 0;
}

int machine_init()
{
  extern int rom_set;

  {
    int have=0;
    FILE *h;
    int i;
    unsigned char rname[5][16]=
    {"tmnt.j17","963-r23","963-u23","963-x23","963-123"};

    printf ("  RomSets you (seem to) have in '%s' :\n     ",rom_path);
    have=0;
    for (i=0;i<5;i++)
    {
      strcpy(temp,rom_path);
      strcat(temp,rname[i]);
      h=fopen(temp,"rb");
      if (h!=NULL) { have+=(1<<i); fclose(h);
        printf ("SET %d  ",i);
      }
    }
    printf ("\n");
    if ((have& (1<<rom_set))==0)
    {
      printf ("You don't seem to have rom_set %d\n",rom_set);
      if (have==0)
      {
        printf ("\nYou don't have any rom sets at all in directory '%s'\n"
        "(Without the original roms any emulator is useless.)\n"
        "This usually means getting the romset, making a subdirectory (folder) %s\n"
        "And then putting the romset into it.\n\n"
        ,rom_path,rom_path);
      }
      else
      {
        int a=0,tmp;
        tmp=have;
        while ((tmp&1)==0) {tmp>>=1; a++;}
        printf ("Since you don't have set %d I'm automatcially selecting set %d\n",rom_set,a);
        rom_set=a;
      }
    }
  }

  printf ("Trying to load ROM SET %d from path '%s'\n",rom_set,rom_path);
  if (rom_set==0)
    printf ("(This rom set looks like TMNT.D05 etc...)\n");
  else if (rom_set==1)
    printf ("(This rom set looks like 963-r21 etc... 4-player)\n");
  else if (rom_set==2)
    printf ("(This rom set looks like 963-u21 etc... 2-player)\n");
  else if (rom_set==3)
    printf ("(This rom set looks like 963-x21 etc... 4-player japan)\n");
  else if (rom_set==4)
    printf ("(This rom set looks like 963-121 etc... 2-player japan)\n");
  else
    printf ("(Not a valid rom set number)\n");



  rom68k=tralloc(0x60000);
  if (rom68k==NULL)
  { printf ("out of memory\n"); exit(1); }
  memset(rom68k,0,0x60000);

  rammem=tralloc(0x04000);
  if (rammem==NULL)
  { printf ("out of memory\n"); exit(1); }
  memset(rammem,0,0x04000);

  colmem=tralloc(0x01000);
  if (colmem==NULL)
  { printf ("out of memory\n"); exit(1); }
  memset(colmem,0,0x01000);

  coomem=tralloc(0x00100);
  if (colmem==NULL)
  { printf ("out of memory\n"); exit(1); }
  memset(coomem,0,0x00100);

  vidmem=tralloc(0x08000);
  if (vidmem==NULL)
  { printf ("out of memory\n"); exit(1); }
  memset(vidmem,0,0x08000);

  sprmem=tralloc(0x00400);
  if (sprmem==NULL)
  { printf ("out of memory\n"); exit(1); }
  memset(sprmem,0,0x00400);

  {
    // nb R2x is 4-player uxx is 2-player
// StarScream specific start!
    // Load the 68k into the start of memory (remembering the byteswap)

    strcpy(temp,rom_path);
    switch(rom_set)
    {
      case 1: strcat(temp,"963-r23"); break;
      case 2: strcat(temp,"963-u23"); break;
      case 3: strcat(temp,"963-x23"); break;
      case 4: strcat(temp,"963-123"); break;
      default: strcat(temp,"tmnt.j17");
    }
    if (load_rom(rom68k+0,temp,2,0x20000)!=0) exit(1);

    strcpy(temp,rom_path);
    switch(rom_set)
    {
      case 1: strcat(temp,"963-r24"); break;
      case 2: strcat(temp,"963-u24"); break;
      case 3: strcat(temp,"963-x24"); break;
      case 4: strcat(temp,"963-124"); break;
      default: strcat(temp,"tmnt.k17");
    }
    if (load_rom(rom68k+1,temp,2,0x20000)!=0) exit(1);

    strcpy(temp,rom_path);
    switch(rom_set)
    {
      case 1: strcat(temp,"963-r21"); break;
      case 2: strcat(temp,"963-u21"); break;
      case 3: strcat(temp,"963-x21"); break;
      case 4: strcat(temp,"963-121"); break;
      default: strcat(temp,"tmnt.j15");
    }
    if (load_rom(rom68k+0x40000,temp,2,0x10000)!=0) exit(1);

    strcpy(temp,rom_path);
    switch(rom_set)
    {
      case 1: strcat(temp,"963-r22"); break;
      case 2: strcat(temp,"963-u22"); break;
      case 3: strcat(temp,"963-x22"); break;
      case 4: strcat(temp,"963-122"); break;
      default: strcat(temp,"tmnt.k15");
    }

    if (load_rom(rom68k+0x40001,temp,2,0x10000)!=0) exit(1);

    byteswap_memory(rom68k,0x60000);
  }

// StarScream specific end!

  z80_mem=tralloc(0x10000);
  if (z80_mem==NULL)
  {
    printf ("out of memory - need 64k for z80\n");
    exit(1);
  }
  memset(z80_mem,0,0x10000);
  strcpy(temp,rom_path);
  switch(rom_set)
  {
    case 1: case 2: case 3: case 4: strcat(temp,"963-e20"); break;
    default: strcat(temp,"tmnt.g13");
  }
  load_rom(z80_mem,temp,1,0x8000);

  {
    // Set up the z80
    static struct mz80context z80={0};
    static struct MemoryReadByte mem_read[]=
    {
      {0x8000,0xffff,z80_read},
      {(UINT32) -1,(UINT32) -1,NULL}
    };
    static struct MemoryWriteByte mem_write[]=
    {
      {0x0000,0x7fff,z80_rom_write},
      {0x8000,0xffff,z80_write},
      {(UINT32) -1,(UINT32) -1,NULL}
    };
    static struct z80PortRead io_read[]=
    {
      {(UINT16) -1,(UINT16) -1,NULL}
    };
    static struct z80PortWrite io_write[]=
    {
      {(UINT16) -1,(UINT16) -1,NULL}
    };
    mz80reset();
    // Modify the default context
    mz80GetContext(&z80);
    z80.z80Base=z80_mem;
    z80.z80MemRead=mem_read;
    z80.z80MemWrite=mem_write;
    z80.z80IoRead=io_read;
    z80.z80IoWrite=io_write;
    mz80SetContext(&z80);
    mz80reset();
  }

  sprite4=tralloc(0x100000);
  if (sprite4==NULL)
  {
    printf ("out of memory - sprite4\n");
    exit(1);
  }

  strcpy(temp,rom_path);
  switch(rom_set)
  {
    case 1: case 2: case 3: case 4: strcat(temp,"963-a17"); break;
    default: strcat(temp,"tmnt.h04");
  }
  load_rom(sprite4+0,temp,3,0x80000); //made these non-fatal

  strcpy(temp,rom_path);
  switch(rom_set)
  {
    case 1: case 2: case 3: case 4: strcat(temp,"963-a15"); break;
    default: strcat(temp,"tmnt.k04");
  }
  load_rom(sprite4+2,temp,3,0x80000);

  printf ("Descrambling sprite4...\n");
  tile_descramble(sprite4,0x100000);

  sprite6=tralloc(0x100000);
  if (sprite6==NULL)
  {
    printf ("out of memory - sprite6\n");
    exit(1);
  }

  strcpy(temp,rom_path);
  switch(rom_set)
  {
    case 1: case 2: case 3: case 4: strcat(temp,"963-a18"); break;
    default: strcat(temp,"tmnt.h06");
  }
  load_rom(sprite6+0,temp,3,0x80000);

  strcpy(temp,rom_path);
  switch(rom_set)
  {
    case 1: case 2: case 3: case 4: strcat(temp,"963-a16"); break;
    default: strcat(temp,"tmnt.k06");
  }
  load_rom(sprite6+2,temp,3,0x80000);

  printf ("\n");
  printf ("Descrambling sprite6...\n");
  tile_descramble(sprite6,0x100000);

  tile27=tralloc(0x100000);
  if (tile27==NULL)
  {
    printf ("out of memory - tile27\n");
    exit(1);
  }

  strcpy(temp,rom_path);
  switch(rom_set)
  {
    case 1: case 2: case 3: case 4: strcat(temp,"963-a28"); break;
    default: strcat(temp,"tmnt.h27");
  }
  load_rom(tile27,temp,3,0x80000);

  strcpy(temp,rom_path); 
  switch(rom_set)
  {
    case 1: case 2: case 3: case 4: strcat(temp,"963-a29"); break;
    default: strcat(temp,"tmnt.k27");
  }
  load_rom(tile27+2,temp,3,0x80000);

  printf ("\n");
  printf ("Descrambling tile27...\n");
  tile_descramble(tile27,0x100000);

// StarScream specific start!
// Create the memory access table
  {
    static struct S68000MEMORYFETCH tmnt_fetch[] =
    {
      {0x000000, 0x05ffff, 0},
      {-1, -1, NULL}
    };
    static struct S68000MEMORYIO tmnt_readbyte[] =
    {
      {0x060000, 0x063fff,NULL         ,NULL },
      {0x000000, 0x05ffff,NULL         ,NULL },
      {0x080000, 0x080fff,NULL         ,NULL },
      {0x100000, 0x107fff,NULL         ,NULL },
      {0x140400, 0x1407ff,NULL         ,NULL },
      {0x0a0000, 0x0a00ff,aoo_readbyte ,NULL },
      {-1, -1, NULL,NULL}
    };
    static struct S68000MEMORYIO tmnt_readword[] =
    {
      {0x060000, 0x063fff,NULL         ,NULL },
      {0x000000, 0x05ffff,NULL         ,NULL },
      {0x080000, 0x080fff,NULL         ,NULL },
      {0x100000, 0x107fff,NULL         ,NULL },
      {0x140400, 0x1407ff,NULL         ,NULL },
      {0x0a0000, 0x0a00ff,aoo_readword ,NULL },
      {-1, -1, NULL,NULL}
    };
    static struct S68000MEMORYIO tmnt_writebyte[] =
    {
      {0x060000, 0x063fff,NULL         ,NULL },
      {0x080000, 0x080fff,NULL         ,NULL },
      {0x0c0000, 0x0c00ff,NULL         ,NULL },
      {0x100000, 0x107fff,NULL         ,NULL },
      {0x140400, 0x1407ff,NULL         ,NULL },
      {0x0a0000, 0x0a00ff,aoo_writebyte,NULL },
      {-1, -1, NULL,NULL}
    };
    static struct S68000MEMORYIO tmnt_writeword[] =
    {
      {0x060000, 0x063fff,NULL         ,NULL },
      {0x080000, 0x080fff,NULL         ,NULL },
      {0x0c0000, 0x0c00ff,NULL         ,NULL },
      {0x100000, 0x107fff,NULL         ,NULL },
      {0x140400, 0x1407ff,NULL         ,NULL },
      {0x0a0000, 0x0a00ff,aoo_writebyte,NULL },
      {-1, -1, NULL,NULL}
    };

    tmnt_fetch[0].off=(unsigned)rom68k;

    tmnt_readbyte[0].userdata =rammem;
    tmnt_readbyte[1].userdata =rom68k;
    tmnt_readbyte[2].userdata =colmem;
    tmnt_readbyte[3].userdata =vidmem;
    tmnt_readbyte[4].userdata =sprmem;

    tmnt_readword[0].userdata =rammem;
    tmnt_readword[1].userdata =rom68k;
    tmnt_readword[2].userdata =colmem;
    tmnt_readword[3].userdata =vidmem;
    tmnt_readword[4].userdata =sprmem;

    tmnt_writebyte[0].userdata =rammem;
    tmnt_writebyte[1].userdata =colmem;
    tmnt_writebyte[2].userdata =coomem;
    tmnt_writebyte[3].userdata =vidmem;
    tmnt_writebyte[4].userdata =sprmem;

    tmnt_writeword[0].userdata =rammem;
    tmnt_writeword[1].userdata =colmem;
    tmnt_writeword[2].userdata =coomem;
    tmnt_writeword[3].userdata =vidmem;
    tmnt_writeword[4].userdata =sprmem;

    s68000context.memoryfetch     = tmnt_fetch;
    s68000context.memoryreadbyte  = tmnt_readbyte;
    s68000context.memoryreadword  = tmnt_readword;
    s68000context.memorywritebyte = tmnt_writebyte;
    s68000context.memorywriteword = tmnt_writeword;
    s68000init();
    s68000reset();
  }
// StarScream specific end!

  do_int5=0;

  cycles=0;
  mega_cycles=0;

  if (!reboot)
  {
    int ret;
    ret=machine_load(autosave_name);
    if (ret==0)
      printf ("LOADED STATE FROM %s",autosave_name);
    else if (ret==-1)
      ;
    else
      printf ("(Error loading state from tmntsave.sav)");
  }
  printf ("\n");

  return 0;
}

int machine_exit()
{
  int total_kilo_cycles;

  if (save_state)
  {
    machine_save(autosave_name);
    printf ("SAVED STATE to %s\n",autosave_name);
  }

  if (tile27!=NULL) tree(tile27);
  if (sprite6!=NULL) tree(sprite6);
  if (sprite4!=NULL) tree(sprite4);
  if (z80_mem!=NULL) tree(z80_mem);

  if (sprmem!=NULL) tree(sprmem);
  if (vidmem!=NULL) tree(vidmem);
  if (coomem!=NULL) tree(coomem);
  if (colmem!=NULL) tree(colmem);
  if (rammem!=NULL) tree(rammem);
  if (rom68k!=NULL) tree(rom68k);


  total_kilo_cycles=mega_cycles*1000+cycles/1000;
  return total_kilo_cycles;
}

