#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dpmi.h>
#include <sys/movedata.h> // for _dosmemputb
#include <sys/segments.h> // for cs
#include <pc.h>
#include "dtmnt.h"
#include <allegro.h>

static int dmabuf_seg=0,dmabuf_sel=0;
int dmabuf_size=4*1024;
int sbsnd_sample_rate=11025;
int sbsnd_stereo16bit=0;
int sbsnd_base=0x220,irq_num=5,dma_chan_8=1,dma_chan_16=5;
static dma_chan=1;
static unsigned int dsp_ver=0;
// Do it EXACTLY like sbdriver.c !
static __dpmi_paddr old_handler, new_handler;
static _go32_dpmi_seginfo wrapper;


// Allocate some DOS memory that doesn't cross a PAGE boundary (for the DMA
// buffer.)
// Returns a segment ((segment<<4)+bytes won't cross)
// And a selector (selector MAY cross) : USE SELECTOR TO DEALLOCATE
static int same_page_dos_mem(int bytes,int *seg, int *sel,int twopage)
{
  int dos_seg=0;
  int hits=0;
  int paras;
  paras=((bytes+15)>>4);
  if (twopage)
  {
    if (paras>0x2000)
    {
      printf ("Can't allocate more than a two page size (%d bytes)\n",paras);
      return 1; // no chance!
    }
  }
  else
  {
    if (paras>0x1000)
    {
      printf ("Can't allocate more than a page size (%d bytes)\n",paras);
      return 1; // no chance!
    }
  }
  dos_seg=__dpmi_allocate_dos_memory(paras,sel);
  if (dos_seg==-1)
  { printf ("Failed to allocate %d bytes DOS memory\n",paras<<4); return 1; }
// NB - units here will be PARAS, not bytes
  // Find out how far up from the start of the block the page boundary hits
  if (twopage)
    hits=0x2000-(dos_seg&0x1fff);
  else
    hits=0x1000-(dos_seg&0x0fff);

  if (hits>=paras) { *seg=dos_seg; return 0; }
  __dpmi_free_dos_memory(*sel);



  paras<<=1;
  // If we allocate twice as much, one half has got to be in!
  dos_seg=__dpmi_allocate_dos_memory(paras,sel);
  if (dos_seg==-1)
  { printf ("Failed to allocate %d bytes DOS memory\n",paras<<4); return 1; }
  // Find out how far up from the start of the block the page hits
  if (twopage)
    hits=0x2000-(dos_seg&0x1fff);
  else
    hits=0x1000-(dos_seg&0x0fff);
  if (hits>=paras) { *seg=dos_seg; return 0; }
  else             { *seg=(dos_seg+hits); return 0; }

  return 0;
}

static int allocate_dma_buffer()
{
  // Make sure divisible by 4
  dmabuf_size&=0xfffffffc;
  if (same_page_dos_mem(dmabuf_size,&dmabuf_seg,&dmabuf_sel,sbsnd_stereo16bit)!=0)
  {
    printf ("Failed to allocate %d byte dma buffer.\n",dmabuf_size);
    return 1;
  }
  return 0;
}

static int deallocate_dma_buffer()
{ __dpmi_free_dos_memory(dmabuf_sel); dmabuf_seg=0; dmabuf_sel=0; return 0; }


static void dsp_in(unsigned int *val)
{ while( !(inportb(sbsnd_base + 0xe) & 0x80)); *val = inportb(sbsnd_base + 0xa); }
static void dsp_out(unsigned int val)
{ while( inportb(sbsnd_base + 0xc) & 0x80); outportb(sbsnd_base + 0xc, val); }

int delay_var=0; // not static - So it doesn't get optimised!
static int reset_dsp(void)
{
  unsigned int val;
  int d=0;
  outportb(sbsnd_base + 0x6, 1);
  for (d=0;d<10000;d++) delay_var+=d;
  outportb(sbsnd_base + 0x6, 0);
  // ------ Check for correct DSP ready response ------
  dsp_in(&val);
  if(val==0xaa)
  {
    int tmp=0;
    // ------ Get DSP version ------
    dsp_out(0x00e1);
    dsp_in(&tmp);
    dsp_ver=(tmp<<8);
    // sbdriver also reads another byte for the sub-version
    dsp_in(&tmp);
    dsp_ver|=tmp;
    return 0;
  }
  printf ("DSP reset (A%x) failed.\n",sbsnd_base);
  return 1;
}

volatile int dma_int_count=0;
void dma_isr(void)
{
  dma_int_count++;
  if (sbsnd_stereo16bit)
  {
    int int_status;
    // Check for correct interrupt status
    outportb(sbsnd_base + 0x4, 0x82);    // selects intr. reg. in mixer
    int_status = inportb(sbsnd_base + 0x5);
    if (int_status & 2)
      inportb(sbsnd_base+0xf);        // Acknowledge interrupt (16-bit)
  }
  else
    inportb(sbsnd_base+0xe);           // Acknowledge interrupt (8-bit)

  if(irq_num>7)                          // End of interrupt
    outportb(0xa0,0x20);
  outportb(0x20,0x20);
}
void dma_isr_end() { }


int lock_dma_isr_code()
{

  if
  (
     _go32_dpmi_lock_code(dma_isr,((char *)dma_isr_end-(char *)dma_isr))
  || _go32_dpmi_lock_data(&wrapper,sizeof(_go32_dpmi_seginfo))
  || _go32_dpmi_lock_data((int *)&dma_int_count,sizeof(int))
  || _go32_dpmi_lock_data(&sbsnd_base,sizeof(int))
  || _go32_dpmi_lock_data(&irq_num,sizeof(int))
  || _go32_dpmi_lock_data(&dma_chan,sizeof(int))
  || _go32_dpmi_lock_data(&dsp_ver,sizeof(int))
  || _go32_dpmi_lock_data(&sbsnd_stereo16bit,sizeof(int))
  )
    return 1;

  return 0;
}

static int int_vect=0;
static unsigned char pic1_default=0,pic2_default=0;
static int set_dma_isr()
{
  unsigned char pic_mask=0;
  int ret=0;

  if (lock_dma_isr_code()!=0)
  {
    printf ("Couldn't lock DMA ISR code/data\n");
    return 1;
  }

// Dave: I haven't a clue exactly what this does, but it looks v.important
// (CreativeLabs do it too)
// I guess it enables the right interrupts
  pic1_default=inportb(0x21);
  pic2_default=inportb(0xA1);

  if(irq_num<8)
  {
    int_vect=irq_num+0x08;
    pic_mask=1<<irq_num;
    pic_mask=~pic_mask;
    outportb(0x21,pic1_default&pic_mask);         // Enable PIC-1's IRQ
  }
  else
  {
    int_vect=irq_num+0x68;
    pic_mask=1<<(irq_num-8);
    pic_mask=~pic_mask;
    outportb(0x21,pic1_default&0xFB);            // Enable IRQ2
    outportb(0xA1,pic2_default&pic_mask);        // As well as PIC-2's IRQ
  }
// Dave: end

  wrapper.pm_offset=(int)dma_isr;
  wrapper.pm_selector=_my_cs();
  _go32_dpmi_allocate_iret_wrapper(&wrapper);
  new_handler.offset32=wrapper.pm_offset;
  new_handler.selector=wrapper.pm_selector;
  ret=__dpmi_get_protected_mode_interrupt_vector(int_vect,&old_handler);
  if (ret!=0)
    printf ("Error saving old_handler (ret is %d)\n",ret);
  ret=__dpmi_set_protected_mode_interrupt_vector(int_vect,&new_handler);
  if (ret!=0)
    printf ("Error setting new_handler (ret is %d)\n",ret);

  return 0;
}

static int unset_dma_isr()
{
  int ret=0;
  ret=__dpmi_set_protected_mode_interrupt_vector(int_vect,&old_handler);
  if (ret!=0)
    printf ("Error restoring old_handler (ret is %d)\n",ret);
// PICs again...
  outportb(0x21,pic1_default);
  outportb(0xA1,pic2_default);
  return 0;
}

// This is fairly blantantly lifted from dmaw32.c!/sb.c in allegro
struct dmaport { unsigned char addr, count, page, mask, mode, FF;};

const struct dmaport dmaport[8] =
{
  {0x00, 0x01, 0x87, 0x0A, 0x0B, 0x0C},  // chan 0
  {0x02, 0x03, 0x83, 0x0A, 0x0B, 0x0C},  // chan 1
  {0x04, 0x05, 0x81, 0x0A, 0x0B, 0x0C},  // chan 2  DON'T USE
  {0x06, 0x07, 0x82, 0x0A, 0x0B, 0x0C},  // chan 3
  {0x00, 0x00, 0x00, 0xD4, 0xD6, 0xD8},  // chan 4  DON'T USE
  {0xC4, 0xC6, 0x8B, 0xD4, 0xD6, 0xD8},  // chan 5
  {0xC8, 0xCA, 0x89, 0xD4, 0xD6, 0xD8},  // chan 6
  {0xCC, 0xCE, 0x8A, 0xD4, 0xD6, 0xD8}   // chan 7
};


static int pagePort[8] = { 0x87, 0x83, 0x81, 0x82, -1, 0x8B, 0x89, 0x8A };
static void sbrips_dma() {
  int offset, page;
  int port;
  int linearAddressOfBuffer=0;
  int length=0;

  length=dmabuf_size-1;
  linearAddressOfBuffer=dmabuf_seg<<4;

  page=linearAddressOfBuffer>>16;
  offset=linearAddressOfBuffer&0xFFFF;

  outportb(0x0A,dma_chan|0x04);              /* Mask DMA channel     */
  outportb(0x0C,0);                             /* Clear byte flip-flop */
  outportb(0x0B,dma_chan|0x58);        /* Set mode             */

  port=dma_chan<<1;
  outportb(port,offset&0xFF);                   /* Program offset       */
  outportb(port,offset>>8);

  port+=1;
  outportb(port,length&0xFF);                   /* Program length       */
  outportb(port,length>>8);

  outportb(pagePort[dma_chan],page);         /* Program page         */

  outportb(0x0A,dma_chan);                   /* Clear channel mask   */
}

static int program_dma_dsp()
{
  struct dmaport dma={0};
  if(dma_chan == 2)
  {
    printf ("DMA channel %d is invalid.\n",dma_chan);
    return 1;
  }
  dma = dmaport[dma_chan];



  // Program DMA controller
// dave added &3
  outportb(dma.mask, (dma_chan | 4));           // disable DMA
  outportb(dma.mode, ((dma_chan&3) | 0x58));        // auto-init
  outportb(dma.FF,  0);                         // clear flip-flop

  {
    unsigned int page=0,offset=0;
    page = (dmabuf_seg>>12);
    if (sbsnd_stereo16bit)
      offset = (dmabuf_seg<<3)&0xFFFF;
    else
      offset = (dmabuf_seg<<4)&0xFFFF;
    outportb(dma.addr, offset&0xff);  // LO byte address of buffer
    outportb(dma.addr, offset>>8  );  // HI byte address of buffer
    outportb(dma.page, page );        // Physical page number
  }

  outportb(dma.FF,  0);                         // clear flip-flop

  if (sbsnd_stereo16bit)
  {
    outportb(dma.count, (((dmabuf_size/2)-1) & 0xFF));// LO byte of count
    outportb(dma.count, (((dmabuf_size/2)-1) >> 8));  // HI byte of count
  }
  else
  {
    outportb(dma.count, ((dmabuf_size-1) & 0xFF));// LO byte of count
    outportb(dma.count, ((dmabuf_size-1) >> 8));  // HI byte of count
  }


// dave added &3
  outportb(dma.mask, dma_chan&3); // enable DMA

// Now GO!
  if (sbsnd_stereo16bit)
  {
    // Needs DSP 4+
    dsp_out(0x41);     // DSP output transfer rate
    dsp_out(sbsnd_sample_rate>> 8); // HI byte
    dsp_out(sbsnd_sample_rate&0xff);// Low byte
    dsp_out(0xb6); // 16-bit auto init (8-bit is 0xc6)
    // 8-bit: 00 = mono, 20 = stereo, 10/30 16-bit
    // I think it's actually |0x10 for signed and |0x20 for stereo
    dsp_out(0x20);
    // ------ Program number of samples to play ------
    dsp_out(((dmabuf_size/4)-1) &0xff); // Low byte
    dsp_out(((dmabuf_size/4)-1) >> 8); // Hi byte
  }
  else
  {
    // Old way of doing it
    unsigned char tc=0;
    // Set transfer time constant (HI-byte only)
    dsp_out(0x0040);
    tc=(256 - 1000000/
      (1 /*2 if stereo*/ * sbsnd_sample_rate));
    dsp_out(tc);

    dsp_out(0x48);
    dsp_out(((dmabuf_size/2)-1) & 0x00FF); // bytes left
    dsp_out(((dmabuf_size/2)-1) >> 8);
    dsp_out(0x001C); // auto-init
  }

  dsp_out(0xd1);  // Turn ON speaker

  return 0;
}

static int halt_dma_dsp()
{
  dsp_out(0xd3);  // Turn OFF speaker
if (sbsnd_stereo16bit)
  dsp_out(0xd9);  // halt 16-bit dma
else
  dsp_out(0xda);  // halt 8-bit dma
  reset_dsp();
  return 0;
}

static unsigned char old_mic_vol=0x44,old_master_vol=0x44,old_wave_vol=0x44;

int sbsnd_mixer_preserve()
{
  // preserve settings
  outportb(sbsnd_base+0x4, 0x0a);
  old_mic_vol=inportb(sbsnd_base+0x5);
  outportb(sbsnd_base+0x4, 0x22);
  old_master_vol=inportb(sbsnd_base+0x5);
  outportb(sbsnd_base+0x4, 0x04);
  old_wave_vol=inportb(sbsnd_base+0x5);
  return 0;
}

int sbsnd_mixer_reset()
{
  outportb(sbsnd_base+0x4, 0x0a);
  outportb(sbsnd_base+0x5, old_mic_vol);
  outportb(sbsnd_base+0x4, 0x22);
  outportb(sbsnd_base+0x5, old_master_vol);
  outportb(sbsnd_base+0x4, 0x04);
  outportb(sbsnd_base+0x5, old_wave_vol);
  return 0;
}

int sbsnd_mixer_set(int master,int volume)
{
// #define mixMSTRVOL      0x22
//#define mixVOCVOL       0x04
//#define mixMICVOL       0x0A
//#define mixOUTFILTER    0x0E

  outportb(sbsnd_base+0x4, 0x0a);
  outportb(sbsnd_base+0x5, 0);

  if (master>=0)
  {
    outportb(sbsnd_base+0x4, 0x22);
    outportb(sbsnd_base+0x5, master);
  }
  if (volume>=0)
  {
    outportb(sbsnd_base+0x4, 0x04);
    outportb(sbsnd_base+0x5, volume);
  }
  return 0;
}

static int sbsnd_init()
{
  if (sbsnd_stereo16bit)
    dma_chan=dma_chan_16;
  else
    dma_chan=dma_chan_8;

  if (allocate_dma_buffer()!=0)
  {
    printf ("allocate_dma_buffer() failed.\n");
    return 1;
  }

  // Put a test sample in the dma buffer
  if (sbsnd_stereo16bit)
  {
    signed short *waveform=NULL;
    int i=0;
    waveform=tralloc(dmabuf_size);
    if (waveform==NULL) {printf ("o o m\n"); exit(1);}
    for (i=0;i<dmabuf_size/2;i++)
      waveform[i]=(((i&255)-128)<<3)+0x8000;
    // NB for w/l middle argument is TRANSFERS, not bytes
    _dosmemputb(waveform,dmabuf_size,dmabuf_seg<<4);
    tree(waveform);
  }
  else
  {
    unsigned char *waveform=NULL;
    int i=0;
    waveform=tralloc(dmabuf_size);
    if (waveform==NULL) {printf ("o o m\n"); exit(1);}
    for (i=0;i<dmabuf_size;i++)
      waveform[i]=((i&255)>>3)+64+32+16;

    // NB for w/l middle argument is TRANSFERS, not bytes
    _dosmemputb(waveform,dmabuf_size,dmabuf_seg<<4);
    tree(waveform);
  }

  if (reset_dsp()!=0)
  {
    printf ("reset_dsp() failed.\n");
    return 1;
  }
  printf ("dsp_ver is %.4x\n",dsp_ver);
  if (dsp_ver<0x400)
    if (sbsnd_stereo16bit)
    {
      printf ("(DSP version can't do 16-bit...)\n");
      sbsnd_stereo16bit=0;
      return 1;
    }

  if (set_dma_isr()!=0)
  {
    printf ("set_dma_isr() failed.\n");
    return 1;
  }
  if (program_dma_dsp()!=0)
  {
    printf ("program_dma_dsp() failed.\n");
    return 1;
  }

  return 0;
}

static int sbsnd_exit()
{
  halt_dma_dsp();
  unset_dma_isr();

  deallocate_dma_buffer();
  return 0;
}

unsigned char *sbsnd_buffer=NULL;

static int last_int_count=0;
int play_sbsnd_stream(int size,int bits,int stereo,int rate,int vol,int pan)
{
  (void)(stereo+vol+pan);
  sbsnd_stereo16bit=0;
  if ((bits==16)^(stereo))
  {
    printf ("Have to have both 16-bit AND stereo with sbsnd\n");
    return 1;
  }

  if ((bits==16)&&stereo) sbsnd_stereo16bit=1;

  sbsnd_sample_rate=rate;
  dmabuf_size=size*2*(sbsnd_stereo16bit?4:1);
  dma_int_count=0;
  last_int_count=0;

  if (sbsnd_init()!=0)
  {
    printf ("sbsnd_init() failed.\n");
    return 1;
  }

  sbsnd_buffer=tralloc(dmabuf_size/2);
  if (sbsnd_buffer==NULL)
  {
    printf ("Couldn't alloc sound buffer.\n");
    return 1;
  }
  return 0;
}

int stop_sbsnd_stream()
{
  if (sbsnd_buffer!=NULL) tree(sbsnd_buffer);
  return (sbsnd_exit());
}

void *get_sbsnd_stream_buffer()
{
  int need_more=0;
  if (dma_int_count!=last_int_count) need_more=1;
  last_int_count=dma_int_count;
  if (need_more) return sbsnd_buffer;
  return NULL;
}

int free_sbsnd_stream_buffer()
{
  int half=0;
  // User has written into sbsnd_buffer, so copy it to the right dma half
  last_int_count=dma_int_count;
  half=dma_int_count&1;
  // 1=playing 2nd half, so copy into first
  if (half)
    dosmemput(sbsnd_buffer,dmabuf_size/2,dmabuf_seg<<4);
  else 
    dosmemput(sbsnd_buffer,dmabuf_size/2,(dmabuf_seg<<4)+(dmabuf_size/2));

  return 0;
}

/*
int main(int argc,char **argv)
{
  if (argc>=2) sbsnd_stereo16bit=atoi(argv[1]);
  if (argc>=3) sbsnd_base=strtol(argv[2],NULL,16);
  if (sbsnd_stereo16bit) dma_chan=5;
  if (argc>=4) dma_chan=atoi(argv[3]);
  if (argc>=5) irq_num=atoi(argv[4]);
  printf ("sbsnd_stereo16bit %d A%x dma%d I%d - command line changes\n",
    sbsnd_stereo16bit,sbsnd_base,dma_chan,irq_num);
  if (sbsnd_init()!=0)
  {
    printf ("sbsnd_init() failed\n");
    return 1;
  }

  while (!kbhit())
    ;
  getxkey();

  sbsnd_exit();

  return 0;
}
*/
