// DTMNT emu - memory module

#include <stdio.h>
#include <allegro.h>
#include "mz80.h"
#include "fm.h"
#include "dtmnt.h"
// DTMNT memory stuff

unsigned char aoo[8]={0};

unsigned char sent_to_a0009=0;
int code_waiting_for_z80=0;

// StarScream specific start!
#include "starcpu.h"

int aoo_readbyte(unsigned a)
{
  if ((a>>8)==0xa00)
  {
    a&=0xfffffe;

    if (a==0xa0000) return aoo[0];
    if (a==0xa0002) return aoo[1];
    if (a==0xa0004) return aoo[2];
    if (a==0xa0006) return aoo[3];
    if (a==0xa0014) return aoo[4];

    if (a==0xa0010) return ~dip[1]; 
    if (a==0xa0012) return ~dip[2];
    if (a==0xa0018) return ~dip[3];
    if (a==0xa0016) return 0xff;


    return 0x00;
  }

// don't have a0000 mem anymore   return MEMB(a);
  return 0;
}

static int next_effect=0;
int display_scodes=0;

void aoo_writebyte(unsigned a,unsigned v)
{
  if ((a&0xfffffe)==0xa0000)
  {
    if (v&0x20) do_int5=1;
    if (v==0) do_int5=0;
    if (v&0x08)
    {
      // Don't actually cause a interrupt (for sound effect), but stack it
      sound_effect[sound_effect_count]=next_effect;
      if (sound_effect_count<(SOUND_EFFECT_LIMIT-1))
        sound_effect_count++;
    }
  }

  if (a==0xa0009)
  {
    if (do_sound)
    {
      // I assume it's going to do v&0x08 next!
      if (scode_remap!=NULL)
        next_effect=scode_remap[v&255];
      else
        next_effect=v;

      if (display_scodes)
      {
        if (
             ((display_scodes==1)&&((v>=0x80)&&(v<0xfe)))
             || (display_scodes==2)
           )
        {
          sprintf(message,"Sound Code 0x%.2x",v);
          message_count=-1;
          mescol=mes_blue;
        }
      }
    }
    return;
  }

// don't have a0000 mem anymore  mem[(a^1)&MEM_MASK]=v;
  return;
}

void aoo_writeword(unsigned a,unsigned v)
{
  (void)(a+v);
  return;
}
int aoo_readword(unsigned a)
{
  (void)(a);
  return 0;
}

// StarScream specific end!



// z80 memory stuff
void z80_rom_write(UINT32 a,UINT8 v,struct MemoryWriteByte *huh)
{
  (void)huh; (void)a; (void)v;
  return;
}

unsigned char booreg[16]={0};

void z80_write(UINT32 a,UINT8 v,struct MemoryWriteByte *huh)
{
  static int set_reg=0;
  (void)huh;
  if (a<0x9000) { z80_mem[a&0xffff]=v; return; }

  if (a==0xc000) {set_reg=v; return;}
  else if (a==0xc001) {OPMWriteReg(0,set_reg,v); return; }

// Just trying to see what else z80 is talking to
//  if (debug_log!=NULL)
//    fprintf(debug_log,"z80_write(%4x,%2x)\n",a,v);

  if (a>=0xb000) if (a<0xb010) booreg[a&15]=v;


  z80_mem[a&0xffff]=v;
  return;
}

UINT8 z80_read(UINT32 a, struct MemoryReadByte *huh)
{
  int drum_addr=0,note=0,vol=0,play=0,channel=0;
  extern int play_a_drum(int drumno,int speed,int vol,int d);

  (void)huh;

  if (a<0x9000) return z80_mem[a&0xffff];

  if (a==0xc001)
  {
    unsigned char tmp=ym2151_timers;
    ym2151_timers=0;
    return tmp;
  }  

  if (a==0xf000) return 1; // change

// Hooked the activation to the Read like JesusLA said
  play=0;
  if (a==0xb005)
  {
    play=1; channel=0;
    drum_addr=(booreg[4]<<16)+(booreg[3]<<8)+booreg[2];
    note=booreg[0]+(booreg[1]<<8);
    vol=booreg[0xc]>>4;
  }
  else if (a==0xb00b)
  {
    play=1; channel=1;
    drum_addr=(booreg[10]<<16)+(booreg[9]<<8)+booreg[8];
    note=booreg[6]+(booreg[7]<<8);
    vol=booreg[0xc]&15;
  }

// Apparently JesusLA says the formula is: 894886.25 / (4096 - sr)

  if (play)
  {
    extern double alter_pitch;
    long long base_freq;
    base_freq=(long long)((894886.0*4.0 + 1.0)*alter_pitch);
    note=(4096-note);
    if (note) note=(base_freq/note)>>2;
    play_a_drum(drum_addr,note,vol,channel);
  }



// Just trying to see what else z80 is talking to
//  if (debug_log!=NULL)
//    fprintf(debug_log,"z80_read(%4x)\n",a);

  if ((a&0xf000)==0xa000)
  {
    if (code_waiting_for_z80)
    {
//      if (debug_log!=NULL) fprintf (debug_log,"z80 asked for this code: %x\n",sent_to_a0009);
      code_waiting_for_z80=0;
      return sent_to_a0009;
    }
    else
    {
      if (debug_log!=NULL) fprintf (debug_log,"z80 asking for code - nothing waiting!\n");
      return 0;
    }
  }

//  if (a==0x9000) return 0;
  return z80_mem[a&0xffff];
}
